#!/usr/bin/env python
# -*- coding: utf-8 -*-
# _version.py
"""Package initialization for DVHA-Stats."""
# Copyright (c) 2020 Dan Cutright
# This file is part of dicompyler-core, relased under a MIT license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-Stats

__author__ = "Dan Cutright"
__email__ = "dan.cutright@gmail.com"
__version__ = "0.1.7"
__release__ = "0.1.7"
__version_info__ = (0, 1, 7)
